@@ -55,6 +55,19 @@ Version 1.0
 Version 1.1
 - implementation change for floats : this change makes existing bases with
 float fields incompatible. Use script conversion_float to upgrade
+Version 1.2
+- method close() forgot to close _id_pos and _del_rows
+- fix a bug for links to other dbs when their path has spaces
+Version 1.3
+- fix a bug in buzhug_algos when selecting by range
+Version 1.4
+- add boolean type
+- default value for fields can be specified in create() and modified by
+  set_default(field_name,default)
+- change conversion of float 0.0 (diidn't preserve order for small positive 
+floats)
+- move methods to manage db information to new module buzhug_info.py
+- add negative indexing. db[-1] returns the last inserted record
 """
 
 import os
@@ -65,6 +78,7 @@ import tokenize
 import re
 import tempfile
 import shutil
+import urllib
 
 import time
 from datetime import date,datetime, time as dtime
@@ -77,8 +91,9 @@ except NameError:
 
 from buzhug_files import *
 import buzhug_algos
+import buzhug_info
 
-version = "1.1"
+version = "1.3"
 
 # ------------------------
 # The following classes are used to modelize the objects in the
@@ -285,17 +300,18 @@ class Base:
 
     BLOCKSIZE = 131072
     
-    types = {} # key = data class name, value = data class
-    file_types = {} # key = data class, value = file class
-    f_encode = {} # key = data class, value = function to_block
-    f_decode = {} # key = data class, value = function from_block
 
     types_map = [ (int,IntegerFile),(float,FloatFile),
             (str,StringFile),(unicode,UnicodeFile),
-            (date,DateFile),(datetime,DateTimeFile), (dtime, TimeFile)]
+            (date,DateFile),(datetime,DateTimeFile), (dtime, TimeFile),
+            (bool,BooleanFile)]
 
     def __init__(self,basename):
         self.name = self.__name__ = basename
+        self.types = {} # key = data class name, value = data class
+        self.file_types = {} # key = data class, value = file class
+        self.f_encode = {} # key = data class, value = function to_block
+        self.f_decode = {} # key = data class, value = function from_block
         self.info_name = os.path.join(basename,'__info__')
         self.pos_name = os.path.join(basename,'__pos__')
         for (c_obj,c_file) in self.types_map:
@@ -317,8 +333,8 @@ class Base:
     def _register_class(self,class_obj,class_file):
         """Register a data type
         class_obj is the data class (eg int)
-        class_file is the class used to manage the file thats hold
-           the data of this class (eg IntegerFile in buzhug_classes)
+        class_file is the class used to manage the file that holds
+           the data of this class (eg IntegerFile in buzhug_files)
         """
         self.types.update({class_obj.__name__:class_obj})
         self.file_types.update({ class_obj:class_file })
@@ -334,7 +350,7 @@ class Base:
 
     def create(self,*fields,**kw):
         """Creates a database instance and returns a reference to it
-        fields are couples (field_name,field_type)
+        fields are tuples (field_name,field_type[,default_value])
         field_name must not begin by _
         field_type can be one of the values in the dictionary self.types
         A keyword 'mode' can be specified:
@@ -357,18 +373,16 @@ class Base:
             else:
                 raise IOError,"Directory %s already exists" %self.name
         self.field_names = [ f[0] for f in fields ]
-        self.fields = dict(fields)
-        for (field_name,field_type) in self.fields.iteritems():
-            self._validate_field(field_name,field_type)
-        # add two internal fields, __id__ and __version__
+        self.fields = dict([(f[0],f[1]) for f in fields])
+        # set general info about field definition
+        buzhug_info.set_info(self,fields)
+        
         self.field_names = ['__id__','__version__'] + self.field_names
         self.fields['__id__'] = int
         self.fields['__version__'] = int
         # create the directory used for the base
         if not os.path.exists(self.name):
             os.mkdir(self.name)
-        # save information in file __info__
-        self._save_info()
         # create index file
         self._id_pos = IntegerFile(self.name,'_id_pos').create()
         # create positions file
@@ -382,22 +396,12 @@ class Base:
         for f in self.field_names:
             self._file[f] = self.file_types[self.fields[f]](self.name,f)
             self._file[f].create()
-        return self
-
-    def _save_info(self):
-        """Save information in the file __info___"""
-        _info = open(self.info_name,'wb')
-        fields = []
-        for k in self.field_names:
-            if isinstance(self.fields[k],Base):
-                fields.append((k,'<base>'+str(self.fields[k].name)))
-            else:
-                fields.append((k,self.fields[k].__name__))
-        _info.write(' '.join(['%s:%s' %(k,v) for (k,v) in fields]))
-        _info.close()
+        # save information in files __info__ and __defaults__
+        buzhug_info.save_info(self)
         # create class for records with all values set
         self._full_rec = makeRecordClass(self,self.record_class,self.field_names)
-        
+        return self
+
     def open(self):
         """Open an existing database and return a reference to it
         Raise IOError if no base is found for the path entered in __init__
@@ -417,9 +421,10 @@ class Base:
         for (k,v) in fields:
             if v.startswith('<base>'):
                 # reference to an external base
-                _db = Base(v[6:]).open()
-                self._register_base(_db)
-                self.fields[k] = _db.types[v[6:]]
+                base_path = urllib.unquote(v[6:])
+                ext_db = Base(base_path).open()
+                self._register_base(ext_db)
+                self.fields[k] = ext_db
             else:
                 self.fields[k] = self.types[v]
         self.field_names = [ k for (k,v) in fields ]
@@ -428,9 +433,8 @@ class Base:
         self.decode = dict([(k,self.f_decode[self.fields[k]]) 
             for k in self.field_names])
         self._open_files()
-        self._del_rows = DeletedRowsFile(self.name,"__del_rows__").open()
-        self._full_rec = makeRecordClass(self,self.record_class,
-            self.field_names)
+        # read default values
+        self.defaults = buzhug_info.read_defaults(self)
         return self
 
     def _open_files(self):
@@ -440,12 +444,17 @@ class Base:
             self._file[f].open()
         self._id_pos = IntegerFile(self.name,'_id_pos').open()
         self._pos = PositionFile(self).open()
+        self._del_rows = DeletedRowsFile(self.name,"__del_rows__").open()
+        self._full_rec = makeRecordClass(self,self.record_class,
+            self.field_names)
 
     def close(self):
         """Close all files"""
         for f in self._file.values():
             f.close()
         self._pos.close()
+        self._id_pos.close()
+        self._del_rows.close()
 
     def destroy(self):
         """Destroy an existing base"""
@@ -454,6 +463,18 @@ class Base:
                 os.remove(os.path.join(dirpath,filename))
         os.rmdir(self.name)
 
+    def set_default(self,field_name,default):
+        """Set a default value for a field"""
+        fields = []
+        for f in self.field_names[2:]:
+            if f==field_name:
+                fields.append((f,self.fields[f],default))
+            elif self.defaults[f] is None:
+                fields.append((f,self.fields[f]))
+            else:
+                fields.append((f,self.fields[f],self.defaults[f]))
+        buzhug_info.set_info(self,fields)
+
     def insert(self,*args,**kw):
         """Public method to insert a record
         Data can be entered as a list of values ordered like in create(),
@@ -474,7 +495,7 @@ class Base:
             raise NameError,"Specifying the __id__ is not allowed"
         if '__version__' in kw.keys():
             raise NameError,"Specifying the __version__ is not allowed"
-        rec = dict([(f,None) for f in self.fields.keys()])
+        rec = dict([(f,self.defaults[f]) for f in self.field_names[2:]])
         for (k,v) in kw.iteritems():
             self._validate(k,v)
             rec[k] = v
@@ -869,7 +890,13 @@ class Base:
         no field is specified"""
         if field_name in self.field_names:
             raise NameError,"Field %s already exists" %field_name
-        self._validate_field(field_name,field_type)
+        field_def = [field_name,field_type]
+        if default is not None:
+            field_def.append(default)
+
+        # validate field and update dictionary defaults
+        buzhug_info.validate_field(self,field_def)
+
         if after is None:
             indx = 2 # insert after __version__
         elif not after in self.field_names:
@@ -883,9 +910,10 @@ class Base:
             self.file_types[self.fields[field_name]](self.name,field_name)
         self._file[field_name].create()
         # populate field file with default value and update position file
-        self._validate(field_name,default)
         self._pos.add_field(field_name,indx,default)
-        self._save_info()
+        buzhug_info.save_info(self)
+        self._full_rec = makeRecordClass(self,self.record_class,
+            self.field_names)
 
     def drop_field(self,field_name):
         """Remove the specified field name"""
@@ -897,22 +925,13 @@ class Base:
             raise ValueError,"Field __version__ can't be removed"
         indx = self.field_names.index(field_name)
         self.field_names.remove(field_name)
+        del self.defaults[field_name]
+        buzhug_info.save_info(self)
         del self.fields[field_name]
         del self._file[field_name]
         self._pos.drop_field(field_name,indx)
-        self._save_info()
-
-    def _validate_field(self,name,typ):
-        if name == '__id__':
-            raise ValueError,'Field name "__id__" is reserved'
-        elif name.startswith('_'):
-            raise ValueError,"Error for %s : names can't begin with _" \
-                % name
-        if typ not in self.types.values():
-            if isinstance(typ,Base):
-                self._register_base(typ)
-            else:
-                raise TypeError,"type %s not allowed (pouet)" %typ    
+        self._full_rec = makeRecordClass(self,self.record_class,
+            self.field_names)
 
     def _validate(self,k,v):
         """Validate the couple key,value"""
@@ -944,6 +963,8 @@ class Base:
 
     def __getitem__(self,num):
         """Direct access by record id"""
+        if num<0:
+            num = len(self)+num
         # first find the line in position file
         block_pos = self._id_pos.get_block_at_pos(5*num)
         if block_pos[0] == '#':
