@@ -282,8 +282,32 @@ class TimeFile(VariableLengthFile):
             return None
         else:
             return dtime(int(block[1:3]),int(block[3:5]),int(block[5:7]))
+
+class BooleanFile(FixedLengthFile):
+
+    block_len = 2
+
+    def to_block(self,value):
+        if value is None:
+            return '!'+chr(0)
+        elif not isinstance(value,bool):
+            raise ValueError,'Bad type : expected bool, got %s %s' \
+                %(value,value.__class__)
+        else:
+            if value:
+                return '-1'
+            else:
+                return '-0'
+    def from_block(self,block):
+        if block[0]=='!':
+            return None
+        elif block == "-0":
+            return False
+        else:
+            return True
     
 import struct
+
 class IntegerFile(FixedLengthFile):
 
     MIDINT = 2**30
@@ -348,6 +372,8 @@ class FloatFile(FixedLengthFile):
         elif not isinstance(value,float):
             raise ValueError,'Bad type : expected float, got %s %s' \
                 %(value,value.__class__)
+        elif value == 0.0:
+            return '-'+chr(128)+chr(0)*8
         else:
             # get mantissa and exponent
             # f = mant*2**exp, 0.5 <= abs(mant) < 1
@@ -364,7 +390,9 @@ class FloatFile(FixedLengthFile):
             return None
         else:
             s = block[1:]
-            if ord(s[0])<128:
+            if ord(s[0])==128:
+                return 0.0
+            elif ord(s[0])<128:
                 # negative number
                 exp = self.offsetneg-struct.unpack('>H',s[:2])[0]
                 mant = struct.unpack('>d',chr(63)+s[2:])[0] - 1.1
@@ -535,10 +563,15 @@ class ExternalFile(FixedLengthFile):
     block_len = 5
     
     def to_block(self,value):
-        v = [ value.__id__ ]
-        return '-'+struct.pack('>i',value.__id__)
+        if value is None:
+            return '!'+chr(0)*4
+        else:
+            v = [ value.__id__ ]
+            return '-'+struct.pack('>i',value.__id__)
 
     def from_block(self,block):
+        if block[0]=='!':
+            return None
         _id = struct.unpack('>i',block[1:])[0]
         try:
             return self.db[_id]
